import logging
logger = logging.getLogger(__name__)
import requests
import tarfile
import os
import yaml
from tqdm.auto import tqdm
import shutil
from os.path import exists
from ..settings import read_value


def plugin_install(args):
    try:
        from ..factory import BADGER_PLUGIN_ROOT
    except Exception as e:
        logger.error(e)
        return

    # We will not make conda as a dependency of Badger
    # This is just a temp solution
    # Should tell the users to use the install command a conda env is needed
    from conda.cli.python_api import run_command, Commands

    hist = {
        'algo': 'algorithms',
        'env': 'environments',
        'ext': 'extensions',
        'intf': 'interfaces'
    }

    identify = {
        'optimize': 'algorithms',
        'Environment': 'environments',
        'Extension': 'extensions',
        'Interface': 'interfaces'
    }

    if args.plugin_type is None:
        print("Please specify further what you wish to install!")
        return

    if args.plugin_type != 'local' and args.plugin_type not in hist:
        print(f"{args.plugin_type} is an invalid option. Choose one of the following:  algo, env, ext, intf, local")
        return

    plugins_url = read_value('BADGER_PLUGINS_URL')

    if args.plugin_specific is None:
        if args.plugin_type == 'local':
            print('Please provide the path to the local tarball for the plugin you wish to install')
            return
        full_word = hist[f'{args.plugin_type}']
        url = f'{plugins_url}/api/{full_word}'
        r = requests.get(url)
        for elem in r.json():
            if exists(f'{BADGER_PLUGIN_ROOT}/{full_word}/{elem}'):
                print(elem, '         (Already installed)')
            else:
                print(elem)
        return

    plugin_path = ''
    tmp_path = os.path.join(BADGER_PLUGIN_ROOT, '.tmp')
    os.makedirs(tmp_path, exist_ok=True)

    if args.plugin_type == 'local':
        if not exists(args.plugin_specific):
            print('This local tarball does not exist!')
            return
        tarname = os.path.basename(os.path.normpath(args.plugin_specific))
        plugin_name = tarname[:-7]
        local_path = os.path.dirname(args.plugin_specific)
        os.chdir(local_path)
        tar = tarfile.open(f'{tarname}', 'r:gz')
        tar.extractall(tmp_path)
        tar.close()

        histog = {}
        os.chdir(f'{tmp_path}/{plugin_name}')
        with open('__init__.py', 'r') as file:
            info = file.read()
        exec(info, histog)
        identifier = list(histog.keys())[-1]
        full_word = identify[identifier]
        plugin_path += f'{BADGER_PLUGIN_ROOT}/{full_word}/{plugin_name}'
        if exists(plugin_path):
            print('This plugin is already installed!')
            return
        shutil.move(f'{tmp_path}/{plugin_name}', f'{BADGER_PLUGIN_ROOT}/{full_word}')
        shutil.rmtree(tmp_path)


    else:
        full_word = hist[f'{args.plugin_type}']
        targz_path = os.path.join(tmp_path, f'{args.plugin_specific}.tar.gz')

        r_d = requests.get(f'{plugins_url}/api/url/{full_word}/{args.plugin_specific}')
        download_url = r_d.text

        r = requests.get(download_url)
        if r.status_code == 200:
            with open(targz_path, 'wb') as f:
                f.write(r.content)
            os.chdir(tmp_path)
            tar = tarfile.open(f'{args.plugin_specific}.tar.gz', 'r:gz')
            plugin_path += f'{BADGER_PLUGIN_ROOT}/{full_word}/{args.plugin_specific}'
            if exists(plugin_path):
                print('This plugin is already installed!')
                return
            print(f'Installing {args.plugin_specific} into {BADGER_PLUGIN_ROOT}/{full_word} ...')
            tar.extractall(f'{BADGER_PLUGIN_ROOT}/{full_word}')
            tar.close()
        else:
            print("The server does not have this plugin!")
            return

    os.chdir(plugin_path)
    with open("configs.yaml", "r") as stream:
        try:
            configs = yaml.safe_load(stream)
        except yaml.YAMLError as e:
            print(e)
        else:
            dependencies = configs['dependencies']
            print('Installing plugin dependencies ...')
            try:
                dependencies.remove('badger-opt')
            except ValueError:
                pass
            for elem in tqdm(dependencies):
                stdout_str, stderr_str, return_code_int = run_command(Commands.INSTALL, ["-y", f'{elem}'])
                if return_code_int != 0:
                    shutil.rmtree(plugin_path)
                    print(stderr_str)
            print('All dependencies successfully installed!')
            print('Plugin installation complete!')



import os
from datetime import datetime
import numpy as np
import yaml


# https://stackoverflow.com/a/39681672/4263605
# https://github.com/yaml/pyyaml/issues/234#issuecomment-765894586
class Dumper(yaml.Dumper):

    def increase_indent(self, flow=False, indentless=False):
        return super(Dumper, self).increase_indent(flow, False)


def ystring(content):
    if content is None:
        return ''

    return yaml.dump(content, Dumper=Dumper, default_flow_style=False, sort_keys=False)


def yprint(content):
    print(ystring(content), end='')


def norm(x, lb, ub):
    return (x - lb) / (ub - lb)


def denorm(x, lb, ub):
    return (1 - x) * lb + x * ub


def config_list_to_dict(config_list):
    if not config_list:
        return {}

    book = {}
    for config in config_list:
        for k, v in config.items():
            book[k] = v

    return book


def load_config(fname):
    configs = None

    if fname is None:
        return configs

    # if fname is a yaml string
    if not os.path.exists(fname):
        try:
            configs = yaml.safe_load(fname)
            # A string is also a valid yaml
            if type(configs) is str:
                raise Exception(
                    f'Error loading config {fname}: file not found')

            return configs
        except yaml.YAMLError:
            raise Exception(f'Error parsing config {fname}: invalid yaml')

    with open(fname, 'r') as f:
        try:
            configs = yaml.safe_load(f)
        except yaml.YAMLError:
            raise Exception(f'Error loading config {fname}: invalid yaml')

    return configs


def merge_params(default_params, params):
    merged_params = None

    if params is None:
        merged_params = default_params
    elif default_params is None:
        merged_params = params
    else:
        merged_params = {**default_params, **params}

    return merged_params


def range_to_str(vranges):
    # Transfer the range list to a string for better printing
    vranges_str = []
    for var_dict in vranges:
        var = next(iter(var_dict))
        vrange = var_dict[var]
        vranges_str.append({})
        vranges_str[-1][var] = f'{vrange[0]} -> {vrange[1]}'

    return vranges_str


class ParetoFront:

    def __init__(self, rules):
        # rules: ['MAXIMIZE', 'MINIMIZE', ...]
        self.rules = (np.array(rules) == 'MINIMIZE') * 2 - 1
        self.dimension = len(rules)
        self.pareto_set = None
        self.pareto_front = None

    def is_dominated(self, candidate):
        # candidate: (x: array-like, y: array-like)
        # First candidate
        if self.pareto_front is None:
            self.pareto_set = np.array(candidate[0]).reshape(1, -1)
            self.pareto_front = np.array(candidate[1]).reshape(1, -1)
            return False

        dmat = (self.pareto_front - candidate[1]) * self.rules > 0
        scores = np.sum(dmat, axis=1)
        if np.sum(scores == 0):  # candidate is dominated
            return True

        # Drop points that are dominated by candidate
        idx_keep = (scores != self.dimension)
        self.pareto_front = np.vstack(
            (self.pareto_front[idx_keep], candidate[1]))
        self.pareto_set = np.vstack((self.pareto_set[idx_keep], candidate[0]))
        return False


def ts_to_str(ts, format='lcls-log'):
    if format == 'lcls-log':
        return ts.strftime('%d-%b-%Y %H:%M:%S')
    elif format == 'lcls-log-full':
        return ts.strftime('%d-%b-%Y %H:%M:%S.%f')
    elif format == 'lcls-fname':
        return ts.strftime('%Y-%m-%d-%H%M%S')
    else:  # ISO format
        return ts.isoformat()


def str_to_ts(timestr, format='lcls-log'):
    if format == 'lcls-log':
        return datetime.strptime(timestr, '%d-%b-%Y %H:%M:%S')
    elif format == 'lcls-log-full':
        return datetime.strptime(timestr, '%d-%b-%Y %H:%M:%S.%f')
    elif format == 'lcls-fname':
        return datetime.strptime(timestr, '%Y-%m-%d-%H%M%S')
    else:  # ISO format
        return datetime.fromisoformat(timestr)


def curr_ts():
    return datetime.now()


def curr_ts_to_str(format='lcls-log'):
    return ts_to_str(datetime.now(), format)


def get_header(routine):
    try:
        obj_names = [next(iter(d))
                        for d in routine['config']['objectives']]
    except:
        obj_names = []
    try:
        var_names = [next(iter(d))
                        for d in routine['config']['variables']]
    except:
        var_names = []
    try:
        if routine['config']['constraints']:
            con_names = [next(iter(d))
                            for d in routine['config']['constraints']]
        else:
            con_names = []
    except:
        con_names = []
    try:
        sta_names = routine['config']['states'] or []
    except KeyError:
        sta_names = []

    return obj_names + con_names + var_names + sta_names


def run_names_to_dict(run_names):
    runs = {}
    for name in run_names:
        tokens = name.split('-')
        year = tokens[1]
        month = tokens[2]
        day = tokens[3]

        try:
            year_dict = runs[year]
        except:
            runs[year] = {}
            year_dict = runs[year]
        key_month = f'{year}-{month}'
        try:
            month_dict = year_dict[key_month]
        except:
            year_dict[key_month] = {}
            month_dict = year_dict[key_month]
        key_day = f'{year}-{month}-{day}'
        try:
            day_list = month_dict[key_day]
        except:
            month_dict[key_day] = []
            day_list = month_dict[key_day]
        day_list.append(name)

    return runs


def convert_str_to_value(str):
    try:
        return int(str)
    except ValueError:
        pass

    try:
        return float(str)
    except ValueError:
        pass

    try:
        return bool(str)
    except ValueError:
        pass

    return str


def parse_rule(rule):
    if type(rule) is str:
        return {
            'direction': rule,
            'filter': 'ignore_nan',
            'reducer': 'percentile_80',
        }

    # rule is a dict
    try:
        direction = rule['direction']
    except:
        direction = 'MINIMIZE'
    try:
        filter = rule['filter']
    except:
        filter = 'ignore_nan'
    try:
        reducer = rule['reducer']
    except:
        reducer = 'percentile_80'

    return {
        'direction': direction,
        'filter': filter,
        'reducer': reducer,
    }



import numpy as np
import logging
logger = logging.getLogger(__name__)
from operator import itemgetter
from .utils import range_to_str, yprint, merge_params, ParetoFront, norm, denorm, \
     parse_rule


def process_raw(raw, rule):
    # filter = rule['filter']
    reducer = rule['reducer']

    if reducer == 'percentile_80':
        return np.percentile(raw, 80)
    elif reducer == 'mean':
        return np.mean(raw)
    elif reducer == 'median':
        return np.median(raw)


def normalize_routine(routine):
    # Sanity check and config normalization
    #
    # routine
    #  name
    #  algo
    #  env
    #  algo_params
    #  env_params
    #  env_vranges: use for vars normalization in config
    #  config: change the var ranges, the obj rules, and the constraints
    env_vranges = routine['env_vranges']
    config = routine['config']

    # Normalize the variables
    for i, var in enumerate(config['variables']):
        if type(var) is str:
            config['variables'][i] = _dict = {}
            _dict[var] = env_vranges[var]
        else:
            var_name = next(iter(var))  # get the only (first) key in the dict
            vrange_vocs = var[var_name]
            vrange_default = env_vranges[var_name]

            if vrange_vocs is None:
                var[var_name] = vrange_default
                continue

            if vrange_vocs[0] < vrange_default[0]:
                logger.warning(
                    f'variable {var_name}: lower limit {vrange_vocs[0]} exceeds the bound, set to the lower bound {vrange_default[0]}')
                lb = vrange_default[0]
            else:
                lb = vrange_vocs[0]
            if vrange_vocs[1] > vrange_default[1]:
                logger.warning(
                    f'variable {var_name}: upper limit {vrange_vocs[1]} exceeds the bound, set to the upper bound {vrange_default[1]}')
                ub = vrange_default[1]
            else:
                ub = vrange_vocs[1]

            if lb >= ub:  # TODO: add logic to deal with the == condition
                raise Exception(
                    f'variable {var_name}: lower limit {lb} must be lower than the upper limit {ub}!')
            var[var_name] = [lb, ub]

    # Normalize the objectives
    for i, obj in enumerate(config['objectives']):
        if type(obj) is str:
            config['objectives'][i] = _dict = {}
            _dict[obj] = 'MINIMIZE'
        else:
            obj_name = next(iter(obj))
            if obj[obj_name] is None:
                obj[obj_name] = 'MINIMIZE'

    # Normalize the constraints
    try:
        _ = config['constraints']
        if not _:  # empty list
            config['constraints'] = None
    except KeyError:
        config['constraints'] = None

    # Normalize the states
    try:
        _ = config['states']
        if not _:  # empty list
            config['states'] = None
    except KeyError:
        config['states'] = None

    # Normalize the domain scaling
    try:
        _ = config['domain_scaling']
        if not _:  # empty dict
            config['domain_scaling'] = None
    except KeyError:
        config['domain_scaling'] = None

    # Remove the additional info
    del routine['env_vranges']

    return routine


def run_routine(routine, skip_review=False, save=None, verbose=2,
                before_evaluate=None, after_evaluate=None,
                env_ready=None, pf_ready=None, states_ready=None):
    # Review the routine
    if not skip_review:
        print('Please review the routine to be run:\n')

        routine_configs_var = routine['config']['variables']
        routine['config']['variables'] = range_to_str(routine_configs_var)
        print('=== Optimization Routine ===')
        yprint(routine)
        print('')

        while True:
            res = input('Proceed ([y]/n)? ')
            if res == 'n':
                return
            elif (not res) or (res == 'y'):
                break
            else:
                print(f'Invalid choice: {res}')

        routine['config']['variables'] = routine_configs_var

    # Save routine if specified
    if save:
        import sqlite3
        from .db import save_routine

        try:
            save_routine(routine)
        except sqlite3.IntegrityError:
            raise Exception(
                f'Routine {routine["name"]} already existed in the database! Please choose another name.')

    # Set up and run the optimization
    from .factory import get_algo, get_env

    # Instantiate the environment
    Environment, configs_env = get_env(routine['env'])
    _configs_env = merge_params(
        configs_env, {'params': routine['env_params']})

    env = instantiate_env(Environment, _configs_env)

    # Start recording the opt logs
    if env.interface:
        env.interface.start_recording()

    if env_ready:
        env_ready(env)

    optimize, configs_algo = get_algo(routine['algo'])

    from .logger import _get_default_logger
    from .logger.event import Events

    # log the optimization progress
    opt_logger = _get_default_logger(verbose)
    var_names = [next(iter(d)) for d in routine['config']['variables']]
    vranges = np.array([d[next(iter(d))]
                        for d in routine['config']['variables']])
    obj_names = [next(iter(d)) for d in routine['config']['objectives']]
    rules = [d[next(iter(d))] for d in routine['config']['objectives']]
    directions = [parse_rule(rule)['direction'] for rule in rules]
    pf = ParetoFront(directions)
    if pf_ready:
        pf_ready(pf)
    if routine['config']['constraints']:
        con_names = [next(iter(d))
                        for d in routine['config']['constraints']]
        thresholds = [d[next(iter(d))]
                        for d in routine['config']['constraints']]
    else:
        con_names = []
        thresholds = []
    try:
        sta_names = routine['config']['states'] or []
    except KeyError:  # this would happen when rerun an old version routine
        sta_names = []

    # Domain scaling for safety
    try:
        configs_scaling = routine['config']['domain_scaling']
    except KeyError:  # this would happen when rerun an old version routine
        configs_scaling = None
    scaling = get_scaling_func(configs_scaling)

    info = {'count': -1}
    # Make a normalized evaluate function

    def evaluate(X):
        Y = []  # objectives
        I = []  # inequality constraints
        E = []  # equality constraints
        Xo = []  # normalized readback of variables

        # Return current state if X is None
        # Do not do the evaluation due to possible high cost
        if X is None:
            _x = np.array(env._get_vars(var_names))
            x = norm(_x, vranges[:, 0], vranges[:, 1])
            X = x.reshape(1, -1)
            return None, None, None, X

        # Perform domain scaling to avoid boundary violations
        X = scaling(X)
        # Double Check if bounds are violated
        if np.max(X) > 1 or np.min(X) < 0:
            logger.warning('proposed trial solution exceeds the bounds, solution has been clipped at bounds!')
            X = np.clip(X, 0, 1)

        for x in X:
            _x = denorm(x, vranges[:, 0], vranges[:, 1])

            # Use unsafe version to support temp vars
            # We have to trust the users...
            env._set_vars(var_names, _x)
            env.vars_changed(var_names, _x)

            _xo = np.array(env._get_vars(var_names), dtype=np.float64)
            xo = norm(_xo, vranges[:, 0], vranges[:, 1])
            Xo.append(xo)

            # Return the readback rather than the values to be set
            if before_evaluate:
                before_evaluate(_xo)

            # Deal with objectives
            obses = []
            obses_raw = []
            for i, obj_name in enumerate(obj_names):
                rule = rules[i]
                rule_dict = parse_rule(rule)
                obs_raw = env.get_obs(obj_name)
                try:
                    obs = float(obs_raw)
                except:  # obs_raw is a list
                    obs = process_raw(obs_raw, rule_dict)
                if rule_dict['direction'] == 'MAXIMIZE':
                    obses.append(-obs)
                else:
                    obses.append(obs)
                obses_raw.append(obs)
            Y.append(obses)
            obses_raw = np.array(obses_raw, dtype=np.float64)

            # Deal with constraints
            # TODO: Check overlapping between objs and cons
            cons_i = []
            cons_e = []
            cons_raw = []
            for i, con_name in enumerate(con_names):
                relation, thres = thresholds[i][:2]
                con = float(env.get_obs(con_name))
                if relation == 'GREATER_THAN':
                    cons_i.append(con - thres)
                elif relation == 'LESS_THAN':
                    cons_i.append(thres - con)
                else:
                    cons_e.append(con - thres)
                cons_raw.append(con)
            if cons_i:
                I.append(cons_i)
            if cons_e:
                E.append(cons_e)
            cons_raw = np.array(cons_raw, dtype=np.float64)

            # Deal with states
            states = []
            for sta_name in sta_names:
                # A state could be an observation or a variable
                try:
                    sta = env.get_obs(sta_name)
                except:
                    sta = env.get_var(sta_name)
                states.append(sta)

            info['count'] += 1
            _idx_x = np.insert(_xo, 0, info['count'])  # keep the idx info

            is_optimal = not pf.is_dominated((_idx_x, obses_raw))
            solution = (_xo, obses_raw, cons_raw, states, is_optimal,
                        var_names, obj_names, con_names, sta_names)
            opt_logger.update(Events.OPTIMIZATION_STEP, solution)

            if after_evaluate:
                after_evaluate(_xo, obses_raw, cons_raw, states)

        Y = np.array(Y)
        if I:
            I = np.array(I)
        else:
            I = None
        if E:
            E = np.array(E)
        else:
            E = None
        Xo = np.array(Xo)

        return Y, I, E, Xo

    # Start the optimization
    print('')
    solution = (None, None, None, None, None,
                var_names, obj_names, con_names, sta_names)
    opt_logger.update(Events.OPTIMIZATION_START, solution)
    try:
        if not callable(optimize):  # doing optimization through extensions
            configs = {
                'routine_configs': routine['config'],
                'algo_configs': merge_params(configs_algo, {'params': routine['algo_params']}),
                'env_configs': _configs_env,
            }
            optimize = optimize.optimize
        else:
            configs = routine['algo_params']

        # Save system states if applicable
        states = env.get_system_states()
        if states_ready and (states is not None):
            states_ready(states)

        optimize(evaluate, configs)
    except Exception as e:
        opt_logger.update(Events.OPTIMIZATION_END, solution)
        raise e
    opt_logger.update(Events.OPTIMIZATION_END, solution)


def instantiate_env(env_class, configs, manager=None):
    from .factory import get_intf  # have to put here to avoid circular dependencies

    # Configure interface
    # TODO: figure out the correct logic
    # It seems that the interface should be given rather than
    # initialized here
    try:
        intf_name = configs['interface'][0]
    except KeyError:
        intf_name = None
    except Exception as e:
        logger.warning(e)
        intf_name = None

    if intf_name is not None:
        if manager is None:
            Interface, _ = get_intf(intf_name)
            intf = Interface()
        else:
            intf = manager.Interface()
    else:
        intf = None

    env = env_class(intf, configs['params'])

    return env

# The following functions are related to domain scaling
# TODO: consider combine them into a class and make it extensible
def list_scaling_func():
    return ['semi-linear', 'sinusoid', 'sigmoid']


def get_scaling_default_params(name):
    if name == 'semi-linear':
        default_params = {
            'center': 0.5,
            'range': 1,
        }
    elif name == 'sinusoid':
        default_params = {
            'center': 0.5,
            'period': 2,
        }
    elif name == 'sigmoid':
        default_params = {
            'center': 0.5,
            'lambda': 8,
        }
    else:
        raise Exception(f'scaling function {name} is not supported')

    return default_params


def get_scaling_func(configs):
    if not configs:  # fallback to default
        configs = {'func': 'semi-linear'}

    name = configs['func']
    params = configs.copy()
    del params['func']

    default_params = get_scaling_default_params(name)
    params = merge_params(default_params, params)

    if name == 'semi-linear':
        center, range = itemgetter('center', 'range')(params)

        def func(X):
            return np.clip((X - center) / range + 0.5, 0, 1)

    elif name == 'sinusoid':
        center, period = itemgetter('center', 'period')(params)

        def func(X):
            return 0.5 * np.sin(2 * np.pi / period * (X - center)) + 0.5

    elif name == 'sigmoid':
        center, lamb = itemgetter('center', 'lambda')(params)

        def func(X):
            return 1 / (1 + np.exp(-lamb * (X - center)))

    # TODO: consider remove this branch since it's useless
    else:
        raise Exception(f'scaling function {name} is not supported')

    return func



import logging
logger = logging.getLogger(__name__)
import os
import sys
import time
import signal
import pandas as pd
from coolname import generate_slug
from ..utils import load_config, merge_params
from ..utils import config_list_to_dict, curr_ts, ts_to_str
from ..core import run_routine as run
from ..core import normalize_routine


def run_n_archive(routine, yes=False, save=False, verbose=2,
                  fmt='lcls-log-full', sleep=0, flush_prompt=False):
    try:
        from ..archive import archive_run
    except Exception as e:
        logger.error(e)
        return

    var_names = [next(iter(d)) for d in routine['config']['variables']]
    obj_names = [next(iter(d)) for d in routine['config']['objectives']]
    if routine['config']['constraints']:
        con_names = [next(iter(d)) for d in routine['config']['constraints']]
    else:
        con_names = []
    try:
        sta_names = routine['config']['states'] or []
    except KeyError:  # this would happen when rerun an old version routine
        sta_names = []
    # Store solutions in a list to avoid global var def
    solutions = []

    status = {}
    status['paused'] = False

    def handler(*args):
        if status['paused'] == True:
            print('')  # start a new line
            if flush_prompt:  # erase the last prompt
                sys.stdout.write('\033[F')
            raise Exception('Optimization run has been terminated!')
        status['paused'] = True

    signal.signal(signal.SIGINT, handler)

    def before_evaluate(vars):
        if status['paused'] == True:
            res = input('Optimization paused. Press Enter to resume or Ctrl/Cmd + C to terminate: ')
            while res != '':
                if flush_prompt:
                    sys.stdout.write('\033[F')
                res = input(f'Invalid choice: {res}. Please press Enter to resume or Ctrl/Cmd + C to terminate: ')
            if flush_prompt:
                sys.stdout.write('\033[F')
        status['paused'] =  False

    def after_evaluate(vars, obses, cons, stas):
        # vars: ndarray
        # obses: ndarray
        # cons: ndarray
        # stas: list
        ts = curr_ts()
        solution = [ts.timestamp(), ts_to_str(ts, fmt)] + list(obses) + list(cons) + list(vars) + stas
        solutions.append(solution)
        # take a break to let the outside signal to change the status
        time.sleep(sleep)

    # Store system states and other stuff
    storage = {'states': None, 'env': None}

    def states_ready(states):
        storage['states'] = states

    def env_ready(env):
        storage['env'] = env

    try:
        run(routine, yes, save, verbose,
            before_evaluate=before_evaluate, after_evaluate=after_evaluate,
            states_ready=states_ready, env_ready=env_ready)
    except Exception as e:
        if str(e) == 'Optimization run has been terminated!':
            logger.info(e)
        else:
            logger.error(e)

    if solutions:  # only save the run when at least one solution has been evaluated
        df = pd.DataFrame(solutions, columns=['timestamp_raw', 'timestamp'] + obj_names + con_names + var_names + sta_names)
        _run = archive_run(routine, df, storage['states'])
        # Try dump the interface logs
        try:
            path = _run['path']
            filename = _run['filename'][:-4] + 'pickle'
            storage['env'].interface.stop_recording(os.path.join(path, filename))
        except:
            pass


def run_routine(args):
    try:
        from ..factory import get_algo, get_env
    except Exception as e:
        logger.error(e)
        return

    try:
        # Get env params
        _, configs_env = get_env(args.env)

        # Get algo params
        _, configs_algo = get_algo(args.algo)

        # Normalize the algo and env params
        params_env = load_config(args.env_params)
        params_algo = load_config(args.algo_params)
    except Exception as e:
        logger.error(e)
        return
    params_env = merge_params(configs_env['params'], params_env)
    params_algo = merge_params(configs_algo['params'], params_algo)

    # Load routine configs
    try:
        configs_routine = load_config(args.config)
    except Exception as e:
        logger.error(e)
        return

    # Compose the routine
    routine = {
        'name': args.save or generate_slug(2),
        'algo': args.algo,
        'env': args.env,
        'algo_params': params_algo,
        'env_params': params_env,
        # env_vranges is an additional info for the normalization
        # Will be removed after the normalization
        'env_vranges': config_list_to_dict(configs_env['variables']),
        'config': configs_routine,
    }

    # Sanity check and config normalization
    try:
        routine = normalize_routine(routine)
    except Exception as e:
        logger.error(e)
        return

    run_n_archive(routine, args.yes, args.save, args.verbose)



from PyQt5.QtWidgets import QVBoxLayout, QHBoxLayout, QPushButton, QWidget, QPlainTextEdit
from PyQt5.QtWidgets import QComboBox, QCheckBox, QStyledItemDelegate, QLabel, QListWidget, QFrame
from .collapsible_box import CollapsibleBox


class BadgerEnvBox(CollapsibleBox):
    def __init__(self, envs=[], parent=None):
        super().__init__(' Environment + VOCS', parent)

        self.envs = envs

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        vbox = QVBoxLayout()

        name = QWidget()
        hbox_name = QHBoxLayout(name)
        hbox_name.setContentsMargins(0, 0, 0, 0)
        lbl = QLabel('Name')
        lbl.setFixedWidth(64)
        self.cb = cb = QComboBox()
        cb.setItemDelegate(QStyledItemDelegate())
        cb.addItems(self.envs)
        cb.setCurrentIndex(-1)
        self.btn_env_play = btn_env_play = QPushButton('Open Playground')
        btn_env_play.setFixedSize(128, 24)
        hbox_name.addWidget(lbl)
        hbox_name.addWidget(cb, 1)
        hbox_name.addWidget(btn_env_play)
        vbox.addWidget(name)

        params = QWidget()
        hbox_params = QHBoxLayout(params)
        hbox_params.setContentsMargins(0, 0, 0, 0)
        lbl_params_col = QWidget()
        vbox_lbl_params = QVBoxLayout(lbl_params_col)
        vbox_lbl_params.setContentsMargins(0, 0, 0, 0)
        lbl_params = QLabel('Params')
        lbl_params.setFixedWidth(64)
        vbox_lbl_params.addWidget(lbl_params)
        vbox_lbl_params.addStretch(1)
        hbox_params.addWidget(lbl_params_col)

        edit_params_col = QWidget()
        vbox_params_edit = QVBoxLayout(edit_params_col)
        vbox_params_edit.setContentsMargins(0, 0, 0, 0)
        self.edit = edit = QPlainTextEdit()
        edit.setFixedHeight(128)
        vbox_params_edit.addWidget(edit)
        hbox_params.addWidget(edit_params_col)
        vbox.addWidget(params)

        # action_bar = QWidget()
        # hbox_action = QHBoxLayout(action_bar)
        # hbox_action.setContentsMargins(0, 0, 0, 0)
        # self.btn_env_play = btn_env_play = QPushButton('Open Playground')
        # btn_env_play.setFixedSize(128, 24)
        # btn_env_play.hide()  # hide for now to prevent confusions
        # hbox_action.addStretch(1)
        # hbox_action.addWidget(btn_env_play)
        # vbox.addWidget(action_bar)

        seperator = QFrame()
        seperator.setFrameShape(QFrame.HLine)
        seperator.setFrameShadow(QFrame.Sunken)
        seperator.setLineWidth(0)
        seperator.setMidLineWidth(0)
        vbox.addWidget(seperator)

        # Variables config
        var_panel = QWidget()
        vbox.addWidget(var_panel, 1)
        hbox_var = QHBoxLayout(var_panel)
        hbox_var.setContentsMargins(0, 0, 0, 0)
        lbl_var_col = QWidget()
        vbox_lbl_var = QVBoxLayout(lbl_var_col)
        vbox_lbl_var.setContentsMargins(0, 0, 0, 0)
        lbl_var = QLabel('Vars')
        lbl_var.setFixedWidth(64)
        vbox_lbl_var.addWidget(lbl_var)
        vbox_lbl_var.addStretch(1)
        hbox_var.addWidget(lbl_var_col)

        edit_var_col = QWidget()
        vbox_var_edit = QVBoxLayout(edit_var_col)
        vbox_var_edit.setContentsMargins(0, 0, 0, 0)
        action_var = QWidget()
        hbox_action_var = QHBoxLayout(action_var)
        hbox_action_var.setContentsMargins(0, 0, 0, 0)
        vbox_var_edit.addWidget(action_var)
        self.btn_all_var = btn_all_var = QPushButton('Check All')
        self.btn_un_all_var = btn_un_all_var = QPushButton('Uncheck All')
        btn_all_var.setFixedSize(96, 24)
        btn_un_all_var.setFixedSize(96, 24)
        self.btn_add_var = btn_add_var = QPushButton('Add')
        btn_add_var.setFixedSize(96, 24)
        btn_add_var.setDisabled(True)
        self.check_only_var = check_only_var = QCheckBox('Show Checked Only')
        check_only_var.setChecked(False)
        hbox_action_var.addWidget(btn_all_var)
        hbox_action_var.addWidget(btn_un_all_var)
        hbox_action_var.addWidget(btn_add_var)
        hbox_action_var.addStretch()
        hbox_action_var.addWidget(check_only_var)
        self.list_var = QListWidget()
        self.list_var.setViewportMargins(2, 2, 17, 2)
        # self.list_var.setFixedHeight(128)
        vbox_var_edit.addWidget(self.list_var)
        # vbox_var_edit.addStretch()
        hbox_var.addWidget(edit_var_col)

        # Objectives config
        obj_panel = QWidget()
        vbox.addWidget(obj_panel, 1)
        hbox_obj = QHBoxLayout(obj_panel)
        hbox_obj.setContentsMargins(0, 0, 0, 0)
        lbl_obj_col = QWidget()
        vbox_lbl_obj = QVBoxLayout(lbl_obj_col)
        vbox_lbl_obj.setContentsMargins(0, 0, 0, 0)
        lbl_obj = QLabel('Objs')
        lbl_obj.setFixedWidth(64)
        vbox_lbl_obj.addWidget(lbl_obj)
        vbox_lbl_obj.addStretch(1)
        hbox_obj.addWidget(lbl_obj_col)

        edit_obj_col = QWidget()
        vbox_obj_edit = QVBoxLayout(edit_obj_col)
        vbox_obj_edit.setContentsMargins(0, 0, 0, 0)
        action_obj = QWidget()
        hbox_action_obj = QHBoxLayout(action_obj)
        hbox_action_obj.setContentsMargins(0, 0, 0, 0)
        vbox_obj_edit.addWidget(action_obj)
        self.btn_all_obj = btn_all_obj = QPushButton('Check All')
        self.btn_un_all_obj = btn_un_all_obj = QPushButton('Uncheck All')
        btn_all_obj.setFixedSize(96, 24)
        btn_un_all_obj.setFixedSize(96, 24)
        self.check_only_obj = check_only_obj = QCheckBox('Show Checked Only')
        check_only_obj.setChecked(False)
        hbox_action_obj.addWidget(btn_all_obj)
        hbox_action_obj.addWidget(btn_un_all_obj)
        hbox_action_obj.addStretch()
        hbox_action_obj.addWidget(check_only_obj)
        self.list_obj = QListWidget()
        # self.list_obj.setFixedHeight(64)
        self.list_obj.setViewportMargins(2, 2, 17, 2)
        vbox_obj_edit.addWidget(self.list_obj)
        # vbox_obj_edit.addStretch()
        hbox_obj.addWidget(edit_obj_col)

        # Constraints config
        con_panel = QWidget()
        vbox.addWidget(con_panel, 1)
        hbox_con = QHBoxLayout(con_panel)
        hbox_con.setContentsMargins(0, 0, 0, 0)
        lbl_con_col = QWidget()
        vbox_lbl_con = QVBoxLayout(lbl_con_col)
        vbox_lbl_con.setContentsMargins(0, 0, 0, 0)
        lbl_con = QLabel('Cons')
        lbl_con.setFixedWidth(64)
        vbox_lbl_con.addWidget(lbl_con)
        vbox_lbl_con.addStretch(1)
        hbox_con.addWidget(lbl_con_col)

        edit_con_col = QWidget()
        vbox_con_edit = QVBoxLayout(edit_con_col)
        vbox_con_edit.setContentsMargins(0, 0, 0, 0)
        action_con = QWidget()
        hbox_action_con = QHBoxLayout(action_con)
        hbox_action_con.setContentsMargins(0, 0, 0, 0)
        vbox_con_edit.addWidget(action_con)
        self.btn_add_con = btn_add_con = QPushButton('Add')
        btn_add_con.setFixedSize(96, 24)
        btn_add_con.setDisabled(True)
        hbox_action_con.addWidget(btn_add_con)
        hbox_action_con.addStretch()
        self.list_con = QListWidget()
        # self.list_con.setFixedHeight(64)
        self.list_con.setViewportMargins(2, 2, 17, 2)
        vbox_con_edit.addWidget(self.list_con)
        # vbox_con_edit.addStretch()
        hbox_con.addWidget(edit_con_col)

        # States config
        sta_panel = QWidget()
        vbox.addWidget(sta_panel, 1)
        hbox_sta = QHBoxLayout(sta_panel)
        hbox_sta.setContentsMargins(0, 0, 0, 0)
        lbl_sta_col = QWidget()
        vbox_lbl_sta = QVBoxLayout(lbl_sta_col)
        vbox_lbl_sta.setContentsMargins(0, 0, 0, 0)
        lbl_sta = QLabel('States')
        lbl_sta.setFixedWidth(64)
        vbox_lbl_sta.addWidget(lbl_sta)
        vbox_lbl_sta.addStretch(1)
        hbox_sta.addWidget(lbl_sta_col)

        edit_sta_col = QWidget()
        vbox_sta_edit = QVBoxLayout(edit_sta_col)
        vbox_sta_edit.setContentsMargins(0, 0, 0, 0)
        action_sta = QWidget()
        hbox_action_sta = QHBoxLayout(action_sta)
        hbox_action_sta.setContentsMargins(0, 0, 0, 0)
        vbox_sta_edit.addWidget(action_sta)
        self.btn_add_sta = btn_add_sta = QPushButton('Add')
        btn_add_sta.setFixedSize(96, 24)
        btn_add_sta.setDisabled(True)
        hbox_action_sta.addWidget(btn_add_sta)
        hbox_action_sta.addStretch()
        self.list_sta = QListWidget()
        # self.list_sta.setFixedHeight(64)
        self.list_sta.setViewportMargins(2, 2, 17, 2)
        vbox_sta_edit.addWidget(self.list_sta)
        # vbox_sta_edit.addStretch()
        hbox_sta.addWidget(edit_sta_col)

        self.setContentLayout(vbox)

    def config_logic(self):
        self.dict_var = {}
        self.dict_obj = {}
        self.dict_con = {}

    def _fit_content(self, list):
        height = list.sizeHintForRow(0) * list.count() + 2 * list.frameWidth() + 4
        height = max(28, min(height, 192))
        list.setFixedHeight(height)

    def fit_content(self):
        return
        self._fit_content(self.list_var)
        self._fit_content(self.list_obj)
        self._fit_content(self.list_con)
        self._fit_content(self.list_sta)



from abc import ABC, abstractmethod
from typing import List
import pickle
from .utils import merge_params, curr_ts


def log(func):

    def func_log(*args, **kwargs):
        if func.__name__ == 'set_value':
            if 'channel' in kwargs.keys():
                channel = kwargs['channel']
            else:
                channel = args[1]
            if 'value' in kwargs.keys():
                value = kwargs['value']
            else:
                value = args[2]
            args[0]._logs.append({
                'timestamp': curr_ts().timestamp(),
                'action': 'set_value',
                'channel': channel,
                'value': value,
            })

            return func(*args, **kwargs)
        elif func.__name__ == 'get_value':
            if 'channel' in kwargs.keys():
                channel = kwargs['channel']
            else:
                channel = args[1]
            value = func(*args, **kwargs)
            args[0]._logs.append({
                'timestamp': curr_ts().timestamp(),
                'action': 'get_value',
                'channel': channel,
                'value': value,
            })

            return value
        else:
            return func(*args, **kwargs)

    return func_log


class Interface(ABC):

    @property
    @abstractmethod
    def name(self) -> str:
        pass

    @abstractmethod
    def __init__(self, params=None):
        self.params = merge_params(self.get_default_params(), params)
        self._logs = []

    # Get the default params of the interface
    @staticmethod
    @abstractmethod
    def get_default_params() -> dict:
        pass

    @abstractmethod
    def get_value(self, channel: str):
        pass

    @abstractmethod
    def set_value(self, channel: str, value):
        pass

    def start_recording(self):
        self._logs = []

    def stop_recording(self, filename):
        # Do not create the pickle file if log is empty
        # This usually happens when the log decorator is not used in the
        # specific interface
        if not self._logs:
            return

        with open(filename, 'wb') as f:
            pickle.dump(self._logs, f)

        self._logs = []

    def get_values(self, channels: List[str]) -> list:
        values = []
        for c in channels:
            values.append(self.get_value(c))

        return values

    def set_values(self, channels: List[str], values: list):
        assert len(channels) == len(
            values), 'Channels and values number mismatch!'

        for idx, c in enumerate(channels):
            self.set_value(c, values[idx])



from PyQt5.QtWidgets import QWidget, QAbstractSpinBox
from PyQt5.QtCore import Qt, QObject, QEvent


def preventAnnoyingSpinboxScrollBehaviour(self, control: QAbstractSpinBox) -> None:
    control.setFocusPolicy(Qt.StrongFocus)
    control.installEventFilter(self.MouseWheelWidgetAdjustmentGuard(control))


class MouseWheelWidgetAdjustmentGuard(QObject):
    def __init__(self, parent: QObject):
        super().__init__(parent)

    def eventFilter(self, o: QObject, e: QEvent) -> bool:
        widget: QWidget = o
        if e.type() == QEvent.Wheel and not widget.hasFocus():
            e.ignore()
            return True
        return super().eventFilter(o, e)



import os
from datetime import datetime
import logging
logger = logging.getLogger(__name__)
import yaml
import sqlite3
from .settings import read_value


# Check badger database root
BADGER_DB_ROOT = read_value('BADGER_DB_ROOT')
if BADGER_DB_ROOT is None:
    raise Exception('Please set the BADGER_DB_ROOT env var!')
elif not os.path.exists(BADGER_DB_ROOT):
    os.makedirs(BADGER_DB_ROOT)
    logger.info(
        f'Badger database root {BADGER_DB_ROOT} created')


def maybe_create_routines_db(func):

    def func_safe(*args, **kwargs):
        db_routine = os.path.join(BADGER_DB_ROOT, 'routines.db')

        con = sqlite3.connect(db_routine)
        cur = con.cursor()

        cur.execute('create table if not exists routine (name not null primary key, config, savedAt timestamp)')

        con.commit()
        con.close()

        return func(*args, **kwargs)

    return func_safe


def maybe_create_runs_db(func):

    def func_safe(*args, **kwargs):
        db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

        con = sqlite3.connect(db_run)
        cur = con.cursor()

        cur.execute('create table if not exists run (id integer primary key, savedAt timestamp, finishedAt timestamp, routine, filename)')

        con.commit()
        con.close()

        return func(*args, **kwargs)

    return func_safe


def filter_routines(records, tags):
    records_filtered = []
    for record in records:
        try:
            _tags = yaml.safe_load(record[1])['config']['tags']
            if tags.items() <= _tags.items():
                records_filtered.append(record)
        except:
            pass

    return records_filtered


@maybe_create_routines_db
def save_routine(routine):
    db_routine = os.path.join(BADGER_DB_ROOT, 'routines.db')

    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    # Insert a record
    cur.execute('insert into routine values (?, ?, ?)',
                (routine['name'], yaml.dump(routine, sort_keys=False), datetime.now()))

    con.commit()
    con.close()


@maybe_create_routines_db
@maybe_create_runs_db
def remove_routine(name, remove_runs=False):
    db_routine = os.path.join(BADGER_DB_ROOT, 'routines.db')

    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    cur.execute(f'delete from routine where name = "{name}"')

    con.commit()
    con.close()

    if remove_runs:
        # Remove all related run records
        db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

        con = sqlite3.connect(db_run)
        cur = con.cursor()

        cur.execute(f'delete from run where routine = "{name}"')

        con.commit()
        con.close()


@maybe_create_routines_db
def load_routine(name):
    db_routine = os.path.join(BADGER_DB_ROOT, 'routines.db')
    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    cur.execute('select * from routine where name=:name', {'name': name})

    records = cur.fetchall()
    con.close()

    if len(records) == 1:
        return yaml.safe_load(records[0][1]), records[0][2]
    elif len(records) == 0:
        # logger.warning(f'Routine {name} not found in the database!')
        return None, None
    else:
        raise Exception(
            f'Multiple routines with name {name} found in the database!')


@maybe_create_routines_db
def list_routine(keyword='', tags={}):
    db_routine = os.path.join(BADGER_DB_ROOT, 'routines.db')
    con = sqlite3.connect(db_routine)
    cur = con.cursor()

    if not tags:
        cur.execute(f'select name, savedAt from routine where name like "%{keyword}%" order by savedAt desc')

        records = cur.fetchall()
        names = [record[0] for record in records]
        timestamps = [record[1] for record in records]
        con.close()
    else:
        cur.execute(f'select name, config, savedAt from routine where name like "%{keyword}%" order by savedAt desc')

        records = cur.fetchall()
        records = filter_routines(records, tags)
        names = [record[0] for record in records]
        timestamps = [record[2] for record in records]
        con.close()

    return names, timestamps


@maybe_create_runs_db
def save_run(run):
    db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    # Insert a record
    routine_name = run['routine']['name']
    run_filename = run['filename']
    timestamps = run['data']['timestamp_raw']
    time_start = datetime.fromtimestamp(timestamps[0])
    time_finish = datetime.fromtimestamp(timestamps[-1])
    cur.execute('insert into run values (?, ?, ?, ?, ?)',
                (None, time_start, time_finish, routine_name, run_filename))

    rid = cur.lastrowid

    con.commit()
    con.close()

    return rid


@maybe_create_runs_db
def get_runs_by_routine(routine: str):
    db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute(f'select filename from run where routine = "{routine}" order by savedAt desc')
    records = cur.fetchall()
    con.close()

    filenames = [record[0] for record in records]

    return filenames


@maybe_create_runs_db
def get_runs():
    db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute(f'select filename from run order by savedAt desc')
    records = cur.fetchall()

    con.close()

    filenames = [record[0] for record in records]

    return filenames


@maybe_create_runs_db
def remove_run_by_filename(name):
    db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute(f'delete from run where filename = "{name}"')

    con.commit()
    con.close()


@maybe_create_runs_db
def remove_run_by_id(rid):
    db_run = os.path.join(BADGER_DB_ROOT, 'runs.db')

    con = sqlite3.connect(db_run)
    cur = con.cursor()

    cur.execute(f'delete from run where id = {rid}')

    con.commit()
    con.close()



from pkg_resources import resource_filename
from PyQt5.QtWidgets import QHBoxLayout, QWidget, QPushButton
from PyQt5.QtWidgets import QLabel
from PyQt5.QtWidgets import QSizePolicy
from PyQt5.QtGui import QIcon
from PyQt5.QtCore import Qt, QSize
from ..windows.settings_dialog import BadgerSettingsDialog


class BadgerStatusBar(QWidget):
    def __init__(self):
        super().__init__()

        self.init_ui()
        self.config_logic()

    def init_ui(self):
        hbox = QHBoxLayout(self)
        hbox.setContentsMargins(0, 0, 0, 0)

        self.summary = summary = QLabel()
        self.btn_settings = btn_settings = QPushButton()
        icon_path = resource_filename(__name__, '../images/gear.png')
        btn_settings.setFixedSize(24, 24)
        btn_settings.setIcon(QIcon(icon_path))
        # btn_settings.setIconSize(QSize(18, 18))
        summary.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        summary.setAlignment(Qt.AlignCenter)
        hbox.addWidget(summary, 1)
        hbox.addWidget(btn_settings)

    def config_logic(self):
        self.btn_settings.clicked.connect(self.go_settings)

    def set_summary(self, text):
        self.summary.setText(text)

    def go_settings(self):
        dlg = BadgerSettingsDialog(self)
        dlg.exec()



import logging
logger = logging.getLogger(__name__)
import shutil
from os.path import exists


def plugin_remove(args):
	try:
		from ..factory import BADGER_PLUGIN_ROOT
	except Exception as e:
		logger.error(e)
		return

	hist = {
		'algo': 'algorithms',
		'env': 'environments',
		'ext': 'extensions',
		'intf': 'interfaces'
	}

	if args.plugin_type is None or args.plugin_specific is None:
		print('Please specify further which plugin you wish to remove!')
		return
	try:
		full_word = hist[f'{args.plugin_type}']
	except KeyError:
		logger.error(f'{args.plugin_type} is not an existing plugin type')
		return

	plugin_path = f'{BADGER_PLUGIN_ROOT}/{full_word}/{args.plugin_specific}'
	if not exists(plugin_path):
		print(f'The plugin {args.plugin_specific} already does not exist!')
		return
	try:
		shutil.rmtree(plugin_path)
	except OSError as e:
		print(f'Error: {plugin_path} : {e.strerror}')
	else:
		print(f'{args.plugin_specific} was removed successfully from {BADGER_PLUGIN_ROOT}/{full_word}')
		print(f'NOTE: The plugin dependencies for {args.plugin_specific} have not been removed')
